## ========== Copyright Header Begin ==========================================
## 
## CrashTest: Fast and Accurate Fault Analysis Platform
## Copyright (C) 2008-2010  University of Michigan
## http://www.eecs.umich.edu/crashtest
## Contributors include Andrea Pellegrini, Valeria Bertacco, Todd Austin, 
## Kypros Constantinides ,Junhao Jiang, Shobana Sudhakar, and Dan Zhang
## 
## The above named program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public
## License version 2 as published by the Free Software Foundation.
## 
## The above named program is distributed in the hope that it will be 
## useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
## 
## You should have received a copy of the GNU General Public
## License along with this work; if not, write to the Free Software
## Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
##
## Andrea Pellegrini <apellegr@umich.edu>
## Valeria Bertacco  <valeria@umich.edu>
## Todd Austin       <austin@umich.edu>
## University of Michigan
## Electrical Engineering and Computer Science Dept.
## 2260 Hayward St.
## Ann Arbor, MI 48109-2121
## 
## ========== Copyright Header End ============================================

########################################################
#|           Name: CrashTest v1.0
#|
#|             By: Andrea Pellegrini
#|         Credit: Dan Zhang, Junhao Jiang
#|  Last Modified: Jun 22, 2010
#|
#|        Purpose: Parses a structural verilog
#|                 file and determines the gate type,
#|                 gate name, and gate outputs.
#| 				   Then injects the type of faults
#| 	               given as input and the number of faults
#|                 in the netlist. 
#|                 This version works with the GTECH
#|                 library but other libraries can be
#|                 used with very little modifications
#|                 to the script.
#|
#|      Used with: CrashTest v.1.0
#|
#|
#|
#|
#|
########################################################
#|
#| The scripts receives as input a verilog file,
#| the name of the top level module where faults
#| will be injected, the type of faults to inject
#| and the numebr of faults to inject. Invoke the
#| script with the option --help to list all allowed
#| options.
#|
########################################################

#!/usr/bin/perl


use strict;
use warnings;
use Data::Dumper;
use Verilog::Netlist;
use Verilog::Getopt;
use English;
use Getopt::Long;
use File::Path;
use File::Basename;
use Sys::Hostname;
use POSIX qw(ceil floor);

### Commandline Options ########################################################

# Keys are options, values are the string returned from myoption=<str>
my %opt_string;

my $help = 0; 									# Help flag
my $opt_verbose = 0; 						# Verbose flag

# base directory of the simulation
$opt_string{input_netlist} = "";  				# Contains the URL of the netlist to 
$opt_string{temporary_directory} = "./parser_out/";		# Directory where all the temporary outputs of the parser will be stored
$opt_string{output_directory} = "../crashtest_injected/";	# Directory where the injected netlist will be stored
$opt_string{output_filename} = "";
$opt_string{library} = "/usr/caen/synopsys-synth-2006.06-sp4/packages/gtech/src_ver/gtech_lib.v";
$opt_string{clock_net} = "";
$opt_string{reset_net} = "";
$opt_string{top_module} = "";
$opt_string{scan_chains} = 32;
$opt_string{stuck_at} = 0;
$opt_string{bridge} = 0;
$opt_string{path_delay} = 0;
$opt_string{transient} = 0;
$opt_string{stuck_open} = 0;
$opt_string{random_seed} = 1;
$opt_string{headers_directory} = "/n/scoobydoo/y/apellegr/OpenSPARCT1.1.7/design/sys/iop/include/";
$opt_string{verilog_directory} = "";
my @clock_nets = ();
my @input_netlist = ();


### Begin Library-Dependent definitions ########################################################

# To use different libraries you need to chnage these parameters:

# This variable is set to match the name of the input that is used as clock for the 
# sequential elements in the design. CP is used for GTECH, for other libraries you 
# need to change it to match your library.
my @clock_ports = ("CP");

# This variable contains the substring common to all components in the library to
# extract information about them and separate them from other components such 
# memory elements.
my @gates_name_filter = ("GTECH");

# This variable contains the name of the flip flops components used in the netlist
my @flip_flops_name_filter = ("GTECH_FD1");

# For each supported Flip flop, indicate which input should be used to insert path-delay and transient fault
my %flip_flops_input_port = ();
$flip_flops_input_port{"GTECH_FD1"} = "D";

# For each supported Flip flop, indicate which input is used as clock signal
my %flip_flops_clock_port = ();
$flip_flops_clock_port{"GTECH_FD1"} = "CP";

# Gates that are part of the Clock Tree: Only few gates are part of the clock gate, depends on the library
my @cells_allowed_in_clock_tree = ("GTECH_NOT", "GTECH_BUF");

### End Library-Dependent definitions ########################################################

# get options from commandline to override defaults if desired
GetOptions
  (
   \%opt_string,
   (
	'top_module=s',
  'input_netlist=s' => \@input_netlist,
  'library=s',
	'temporary_directory=s',
	'output_directory=s',
	'output_filename=s',
	'clock_net=s' => \@clock_nets,
	'reset_net=s',
	'scan_chains=s',
	'stuck_at=s', 
	'bridge=s',
	'path_delay=s', 
	'transient=s', 
	'stuck_open=s',
	'random_seed=s',
	'headers_directory=s', 
	'verilog_directory=s',
   ),
  'help|h|?'       => \$help,
	'verbose|v|?'    => \$opt_verbose
  );

# Parse basic options
if($help) {
	&usage();
	exit(0);
}


# Check if the library file is proper:
if($opt_string{library} =~ /^$/) {
	&usage();
	exit(1);
} else {
	&can_read_file($opt_string{library});
}

# Check if the input files are proper:
if($#input_netlist == -1) {
        &usage();
        exit(1);
}
foreach(@input_netlist) {
	&can_read_file($_)
}

# Initialize the random seed:
srand($opt_string{random_seed});

# Keep track of the variable that changed and need to be replaced in the rhs of continuous assign statements
my %assign_rhs = ();


print "File: ".$opt_string{input_netlist}." Library: ".$opt_string{library}."\n" if ($opt_verbose);

# If the option are properly set, create the directory where the outputs will be stored
mkdir $opt_string{temporary_directory}, 0777 unless -d $opt_string{temporary_directory};


my $opt = new Verilog::Getopt;
$opt->parameter( "+incdir+", $opt_string{headers_directory}."+".$opt_string{verilog_directory},
	"-y",$opt_string{headers_directory},
);

### Instantiate netlist parser and parse input file ########################################################
my $nl = new Verilog::Netlist ( 
				options => $opt,
				keep_comments => 1,
				use_vars => 1,
				link_read_nonfatal => 1,
				include_open_nonfatal => 1,
);

$nl->read_file (filename=>$opt_string{library});

foreach(@input_netlist) {
	$nl->read_file (filename=>$_);
}

# Process our modules
print "link and lint\n" if($opt_verbose);
$nl->link();
$nl->lint();
$nl->exit_if_error();

# Variable that contains the signals in the clock tree
my @clock_tree_nets = ();

print "Found top level modules in verilog:\n" if($opt_verbose);
foreach my $top_module ($nl->top_modules_sorted) {

	if($top_module->name eq $opt_string{top_module}) {		
		print "\t", $top_module->name, "\n" if($opt_verbose);

		# Extract nets in the clock tree
		my @possible_cells_in_clock_tree = ();
		foreach my $cell ($top_module->cells) {
			foreach my $possible_clock_tree_cell (@cells_allowed_in_clock_tree) {
				if($possible_clock_tree_cell eq $cell->submodname) {
					push(@possible_cells_in_clock_tree, $cell->name);
				}
			}
		}

		my $clock_file_name = $opt_string{temporary_directory}.$top_module->name."_CLOCK.ct";
		# Check if in the temporary directory there is the list of networks in the clock tree:
		if (-e $clock_file_name) { 
			open CLOCK_NET_FILE, $clock_file_name;
			
			while(<CLOCK_NET_FILE>) {
				my $net_name = $_;
				chomp($net_name);
				push(@clock_tree_nets, $net_name);
			}

			close CLOCK_NET_FILE;
		}
		else { 
			while ($#clock_nets >= 0) {
				my $clock_net = pop(@clock_nets);
				print "Analyzing clock tree starting from: ".$clock_net."\n" if($opt_verbose);
				print "Clock signals to analyze: ".$#clock_nets."\n" if($opt_verbose);
				push(@clock_tree_nets, $clock_net);

				my $net = $top_module->find_net($clock_net);

				foreach my $cell_name (@possible_cells_in_clock_tree) {
					my $cell = $top_module->find_cell($cell_name);
					
					my $cell_part_of_clock_tree = 0;
					foreach my $cell_pin ($cell->pins) {
						my $cell_port = $cell_pin->port;

						if(($cell_port->direction eq "in")&&($cell_pin->netname eq $clock_net)) {	
							print "Clock ".$clock_net." is input\n"  if($opt_verbose);			
							print "Clock sent to port ".$cell_port->name."\n"  if($opt_verbose);
							$cell_part_of_clock_tree = 1;					
						}
					}
					if($cell_part_of_clock_tree) {
						foreach my $cell_pin ($cell->pins) {
							my $cell_port = $cell_pin->port;
		
							if($cell_port->direction eq "out") {
								my $new_clock_net = $cell_pin->netname;
								print "Pin ".$cell_pin->name." is part of the clock tree\n" if($opt_verbose);
								print "Net ".$new_clock_net." is part of the clock tree\n" if($opt_verbose);
								push(@clock_nets, $new_clock_net);
								push(@clock_tree_nets, $new_clock_net);
							}
						}
					}
				}
			}
			
			# Store the list of clock nets in a temporary file for future uses:
			open CLOCK_NET_FILE, ">".$clock_file_name;
			foreach(@clock_tree_nets) {
				my $net_name = $_;
				print CLOCK_NET_FILE $net_name."\n";
			}
			close CLOCK_NET_FILE;
	
		}



		my @file_additions = ();

		my $total_faults_locations = 0;

		if($opt_string{stuck_at} > 0) {

			my $num_fault_locations = $opt_string{stuck_at};			

			# Build the list of nodes where a stuck-at fault can be injected. To each net is attached the probability that 
			# a fault hits the network. 

			my %node_list = &return_valid_nets_for_faults();

# %net_list_for_stuck_at

			

			if($opt_verbose) {
				print "Nets where I can insert stuck-at faults:\n";
				for my $net ( keys %node_list ) {
					my $fault_probability = $node_list{$net};
					print "$net => $fault_probability\n";
				}
			}

			# Call the function to select which faulty nets will be instrumented
			my @faulty_nets = &select_faults_from_list(\%node_list, $num_fault_locations);

			if($opt_verbose) {
				print "Nets selected to inject stuck-at faults:\n";
				foreach (@faulty_nets) {
					print $_."\n";
				}
			}

			# Compute the depth of the scan chains:
			my $number_scan_chains = $opt_string{scan_chains};
			my $number_faulty_nets = $#faulty_nets+1;
			my $depth_scan_chains = ceil($number_faulty_nets/$number_scan_chains) - 1;

			$total_faults_locations = $number_faulty_nets;

			push(@file_additions, "// ############################# CRASHTEST ###################################:\n");
			push(@file_additions, "// Declaration signals needed for Crashtest:\n");
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fe_sc_".$scan_chain_index.";\n");
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fv_sc_".$scan_chain_index.";\n");
			}
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "wire [".$depth_scan_chains.":0] crashtest_stuck_at_".$scan_chain_index.";\n");
			}

			print "I will insert ".$number_faulty_nets." faults in ".$number_scan_chains." scan chains each of them deep ".$depth_scan_chains."\n" if($opt_verbose);


			my $fault_index = 0;
			print "Nets selected to inject stuck-at faults:\n" if($opt_verbose);
			foreach (@faulty_nets) {

				my $faulty_net = $_;

				my $scan_chain_selected = &get_scan_chain_for_fault($fault_index);
				my $scan_chain_depth = &get_depth_for_fault($fault_index);
				
				# If the net is an input change it right away, if the net is output of a gate change the output
				my $name_signal = $faulty_net;
				my $index_signal = -1;
				
				if($faulty_net =~ m/(.*)\[(\d*)\]/) {
					$name_signal = $1;
					$index_signal = $2;
				}

				$name_signal =~ s/\\//g;
				$name_signal =~ s/\//\_/g;

				print "Faulty net selected: ".$faulty_net."\n" if($opt_verbose);
				my $driver_found = 0;
				my $original_net = $top_module->find_net($faulty_net);

				# Extract from which cell the signal is generated:					
				foreach my $cell ($top_module->cells) {
					foreach my $cell_pin ($cell->pins) {
						my $cell_port = $cell_pin->port;
						if(defined($cell_port)) {
							if(($cell_port->direction eq "out")&&($cell_pin->netname eq $faulty_net)) {
								print $cell->dump if($opt_verbose);
								print "Found Driver cell ".$cell->name." for selected faulty network ".$faulty_net."\n"  if($opt_verbose);
								$driver_found = 1;

								push(@file_additions, "\n// Fault Number: $fault_index; Faulty Net: $faulty_net; Scan Chain: $scan_chain_selected; Depth: $scan_chain_depth\n");

								if($scan_chain_depth == 0) {
									push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
									push(@file_additions, "begin\n");
									push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[0] <= crashtest_fe_data[".$scan_chain_selected."];\n");
									push(@file_additions, "end\n");

									push(@file_additions, "always@(posedge crashtest_fv_clk)\n");
									push(@file_additions, "begin\n");
									push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."[0] <= crashtest_fv_data[".$scan_chain_selected."];\n");
									push(@file_additions, "end\n");
								} else {
									push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
									push(@file_additions, "begin\n");
									push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= crashtest_fe_sc_".$scan_chain_selected."[".($scan_chain_depth-1)."];\n");
									push(@file_additions, "end\n");

									push(@file_additions, "always@(posedge crashtest_fv_clk)\n");
									push(@file_additions, "begin\n");
									push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= crashtest_fv_sc_".$scan_chain_selected."[".($scan_chain_depth-1)."];\n");
									push(@file_additions, "end\n");

								}

								my $faulty_net_input = "crashtest_stuck_at_".$scan_chain_selected."\[".$scan_chain_depth."\]";
	

								push(@file_additions, "assign ".$faulty_net." = crashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] ? crashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] : $faulty_net_input;\n");

								my @cell_pin_list = $cell->pins_sorted;

								my %new_cell_ports = ();
								my %new_cell_pins = ();

								
								for(my $pin_index = 0; $pin_index <= $#cell_pin_list; $pin_index++) {
									my $cell_port = $cell_pin_list[$pin_index]->port;

									$new_cell_pins{$cell_port->name} = $cell_pin_list[$pin_index]->name;
									if($cell_pin->name eq $cell_port->name) {
										# Change the signal with the new one
										$new_cell_ports{$cell_port->name} = $faulty_net_input;
										
									} else {
										$new_cell_ports{$cell_port->name} = $cell_pin_list[$pin_index]->netname;
									}
								}

								my $cell_name = $cell->name;
								my $cell_module_name = $cell->submodname;
								my $cell_module = $nl->find_module($cell_module_name);
								$cell->delete;

								my $new_cell_object = $top_module->new_cell(name=>$cell_name, submodname=>$cell_module_name,
									filename=>$top_module->filename, lineno=>$top_module->lineno);


								
								for my $key ( keys %new_cell_ports ) {
									my $cell_port = $cell_module->find_port($key);

									$new_cell_object->new_pin(name=>$new_cell_pins{$key}, netname=>$new_cell_ports{$key}, port=>$cell_port,
									filename=>$top_module->filename, lineno=>$top_module->lineno);
								}

								$new_cell_object->lint();

							}
						}
					}
				}

				if(!$driver_found) {
					# If the signal is not generated from a cell, look if the signal is an input of the top module:
					$driver_found = 1;

					my $faulty_net_input = "crashtest_".&trim($name_signal)."_".$fault_index;
					if($index_signal > 0) {
						my $faulty_net_input = "crashtest_".&trim($name_signal)."_".$index_signal."_".$fault_index;
					}

					# If the net is a primary input, I will need to change all the connections to the port: create a new wire to connect the mux to the gates that use it as input
										
					$top_module->new_net(name=>$faulty_net_input,
						data_type=>"wire", 
						comment=>"// Crashtest: Wire needed to insert fault in input signals",);

					push(@file_additions, "\n// Fault Number: $fault_index; Faulty Net: $faulty_net; Scan Chain: $scan_chain_selected; Depth: $scan_chain_depth\n");
				
					if($scan_chain_depth == 0) {
						push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[0] <= crashtest_fe_data[".$scan_chain_selected."];\n");
						push(@file_additions, "end\n");

						push(@file_additions, "always@(posedge crashtest_fv_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."[0] <= crashtest_fv_data[".$scan_chain_selected."];\n");
						push(@file_additions, "end\n");
					} else {
						push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= crashtest_fe_sc_".$scan_chain_selected."[".($scan_chain_depth-1)."];\n");
						push(@file_additions, "end\n");

						push(@file_additions, "always@(posedge crashtest_fv_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= crashtest_fv_sc_".$scan_chain_selected."[".($scan_chain_depth-1)."];\n");
						push(@file_additions, "end\n");

					}

					push(@file_additions, "assign ".$faulty_net_input." = crashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] ? crashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] : ".$faulty_net.";\n");




					# Find all the cells that use the original signal as input and change their connections	
					foreach my $cell ($top_module->cells) {
						foreach my $cell_pin ($cell->pins) {
							my $cell_port = $cell_pin->port;
							if(defined($cell_port)) {
								if(($cell_port->direction eq "in")&&($cell_pin->netname eq $faulty_net)) {
									print $cell->dump if($opt_verbose);
									print "Found cell ".$cell->name." that uses ".$faulty_net." as input.\n" if($opt_verbose);

									my @cell_pin_list = $cell->pins_sorted;

									my %new_cell_ports = ();
									my %new_cell_pins = ();
									for(my $pin_index = 0; $pin_index <= $#cell_pin_list; $pin_index++) {
										my $cell_port = $cell_pin_list[$pin_index]->port;

										$new_cell_pins{$cell_port->name} = $cell_pin_list[$pin_index]->name;
										if($cell_pin->name eq $cell_port->name) {
											# Change the signal with the new one
											$new_cell_ports{$cell_port->name} = $faulty_net_input;
										
										} else {
											$new_cell_ports{$cell_port->name} = $cell_pin_list[$pin_index]->netname;
										}
										if($pin_index < $#cell_pin_list) {

										}
									}

									my $cell_name = $cell->name;
									my $cell_module_name = $cell->submodname;
									my $cell_module = $nl->find_module($cell_module_name);
									$cell->delete;

									my $new_cell_object = $top_module->new_cell(name=>$cell_name, submodname=>$cell_module_name,
										filename=>$top_module->filename, lineno=>$top_module->lineno);


								
									for my $key ( keys %new_cell_ports ) {
										my $cell_port = $cell_module->find_port($key);

										$new_cell_object->new_pin(name=>$new_cell_pins{$key}, netname=>$new_cell_ports{$key}, port=>$cell_port,
										filename=>$top_module->filename, lineno=>$top_module->lineno);
									}

								
								
									$new_cell_object->lint();

								}
							}
						}
					}

					# I need to change all the assign statements to make them use the modified net:
					print "Analyze statements:\n" if($opt_verbose);
					foreach my $statement ($top_module->statements_sorted) {
						print "LHS: ".$statement->lhs if($opt_verbose);
						print " RHS: ".$statement->rhs."\n" if($opt_verbose);

						if($statement->rhs eq $faulty_net) {
							$assign_rhs{$statement->rhs} = $faulty_net_input;
							print "Substituition ".$statement->rhs." to ".$faulty_net_input." annotated!\n" if($opt_verbose);
						}
					}
				}

				$fault_index++;
			}
		} elsif($opt_string{bridge} > 0) {
			my $num_fault_locations = $opt_string{bridge};			

			# Build the list of cells where a bridge fault can be injected to the inputs. To each net is attached the probability that 
			# a fault hits the gate. 

			my %gates_list = &get_list_gates_at_least_two_inputs();

			my @gates_faulty = &select_faults_from_list(\%gates_list, $num_fault_locations);
			
			if($opt_verbose) {
				print "Gates where I will insert bridge faults:\n";
				foreach(@gates_faulty) {
					my $cell = $_;
					print "$cell\n";
				}
			}

			# Compute the depth of the scan chains:
			my $number_scan_chains = $opt_string{scan_chains};
			my $number_faulty_nets = $#gates_faulty+1;
			my $depth_scan_chains = ceil($number_faulty_nets/$number_scan_chains) - 1;

			$total_faults_locations = $number_faulty_nets;

			push(@file_additions, "// ############################# CRASHTEST ###################################:\n");
			push(@file_additions, "// Declaration signals needed for Crashtest:\n");
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fe_sc_".$scan_chain_index.";\n");
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fv_sc_".$scan_chain_index.";\n");
			}
			print "I will insert ".$number_scan_chains." scan chains each of them deep ".$depth_scan_chains."\n" if($opt_verbose);

			my $fault_index = 0;
			foreach my $faulty_cell_name (@gates_faulty) {

				my $scan_chain_selected = &get_scan_chain_for_fault($fault_index);
				my $scan_chain_depth = &get_depth_for_fault($fault_index);

				my @cell_ports_for_bridge = ();
				my @cell_nets_for_bridge = ();

				my $cell = $top_module->find_cell($faulty_cell_name);
				foreach my $cell_pin ($cell->pins) {
					my $cell_port = $cell_pin->port;
					if(defined($cell_port)) {
						my $net_part_of_clock_tree = 0;
						if(($cell_port->direction eq "in")) {
							foreach(@clock_tree_nets) {
								if($cell_pin->netname eq $_) {
									$net_part_of_clock_tree = 1;
								}
							}
							if(!$net_part_of_clock_tree) {
								push(@cell_ports_for_bridge, $cell_port->name);
								push(@cell_nets_for_bridge, $cell_pin->netname);
							}
						}
					}
				}
		
				my $num_input_ports = $#cell_ports_for_bridge + 1;

				if($num_input_ports < 2) {	
					print "I cannot insert bridge fault in a gate with less than 2 not clocked inputs!\n";
					exit(1);
				}

				my $random_index_a = floor(rand($num_input_ports));
				my $random_index_b = floor(rand($num_input_ports));
				while($random_index_a == $random_index_b) {
					$random_index_b = floor(rand($num_input_ports));
				}

				my $faulty_net_a = $cell_nets_for_bridge[$random_index_a];
				my $faulty_net_b = $cell_nets_for_bridge[$random_index_b];

				my $name_signal_a = $faulty_net_a;
				my $index_signal_a = -1;
				
				if($faulty_net_a =~ m/(.*)\[(\d*)\]/) {
					$name_signal_a = $1;
					$index_signal_a = $2;
				}

				$name_signal_a =~ s/\\//g;
				$name_signal_a =~ s/\//\_/g;

				print "Signal A name: ".$name_signal_a." index: ".$index_signal_a."\n" if($opt_verbose);
				
				my $name_signal_b = $faulty_net_b;
				my $index_signal_b = -1;
				
				if($faulty_net_b =~ m/(.*)\[(\d*)\]/) {
					$name_signal_b = $1;
					$index_signal_b = $2;
				}

				$name_signal_b =~ s/\\//g;
				$name_signal_b =~ s/\//\_/g;

				print "Signal B name: ".$name_signal_b." index: ".$index_signal_b."\n" if($opt_verbose);

				my $faulty_net_input_a = &trim($faulty_net_a);
				
				my $faulty_net_output_a = "crashtest_bridge_".&trim($name_signal_a)."_".$fault_index;
				if($index_signal_a >= 0) {
					$faulty_net_output_a = "crashtest_bridge_".&trim($name_signal_a)."_".$index_signal_a."_".$fault_index;
				}

				$top_module->new_net(name=>$faulty_net_output_a,
					data_type=>"wire", 
					comment=>"// Crashtest: Wire needed to insert fault in input signals",);

				my $faulty_net_input_b = $faulty_net_b;

				my $faulty_net_output_b = "crashtest_bridge_".&trim($name_signal_b)."_".$fault_index;
				if($index_signal_b >= 0) {
					$faulty_net_output_b = "crashtest_bridge_".&trim($name_signal_b)."_".$index_signal_b."_".$fault_index;
				}

				$top_module->new_net(name=>$faulty_net_output_b,
					data_type=>"wire", 
					comment=>"// Crashtest: Wire needed to insert fault in input signals",);

				my @cell_pin_list = $cell->pins_sorted;

				my %new_cell_ports = ();
				my %new_cell_pins = ();
				for(my $pin_index = 0; $pin_index <= $#cell_pin_list; $pin_index++) {
					my $cell_port = $cell_pin_list[$pin_index]->port;
					my $cell_pin = $cell_pin_list[$pin_index];

					$new_cell_pins{$cell_port->name} = $cell_pin_list[$pin_index]->name;
					if($cell_pin->name eq $cell_ports_for_bridge[$random_index_a]) {
						# Change the signal with the new one
						$new_cell_ports{$cell_port->name} = $faulty_net_output_a;					
					} elsif($cell_pin->name eq $cell_ports_for_bridge[$random_index_b]) {
						$new_cell_ports{$cell_port->name} = $faulty_net_output_b;	
					} else {
						$new_cell_ports{$cell_port->name} = $cell_pin_list[$pin_index]->netname;
					}
				}

				my $cell_name = $cell->name;
				my $cell_module_name = $cell->submodname;
				my $cell_module = $nl->find_module($cell_module_name);
				$cell->delete;

				my $new_cell_object = $top_module->new_cell(name=>$cell_name, submodname=>$cell_module_name,
					filename=>$top_module->filename, lineno=>$top_module->lineno);
			
				for my $key ( keys %new_cell_ports ) {
					my $cell_port = $cell_module->find_port($key);

					$new_cell_object->new_pin(name=>$new_cell_pins{$key}, netname=>$new_cell_ports{$key}, port=>$cell_port,
					filename=>$top_module->filename, lineno=>$top_module->lineno);
				}

			
			
				$new_cell_object->lint();

				print "Faulty nets selected to bridge: ".$faulty_net_a." <--> ".$faulty_net_b."\n" if($opt_verbose);

				push(@file_additions, "\n// Fault Number: $fault_index; Faulty Net A: $faulty_net_a; Faulty Net B: $faulty_net_b; Scan Chain: $scan_chain_selected; Depth: $scan_chain_depth\n");
			
				if($scan_chain_depth == 0) {
					push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
					push(@file_additions, "begin\n");
					push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."\[0\] <= crashtest_fe_data\[".$scan_chain_selected."\];\n");
					push(@file_additions, "end\n");

					push(@file_additions, "always@(posedge crashtest_fv_clk)\n");
					push(@file_additions, "begin\n");
					push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."\[0\] <= crashtest_fv_data\[".$scan_chain_selected."\];\n");
					push(@file_additions, "end\n");
				} else {
					push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
					push(@file_additions, "begin\n");
					push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] <= crashtest_fe_sc_".$scan_chain_selected."\[".($scan_chain_depth-1)."\];\n");
					push(@file_additions, "end\n");

					push(@file_additions, "always@(posedge crashtest_fv_clk)\n");
					push(@file_additions, "begin\n");
					push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] <= crashtest_fv_sc_".$scan_chain_selected."\[".($scan_chain_depth-1)."\];\n");
					push(@file_additions, "end\n");
				}

				# Connect the selected nets to the multiplexers and the majority voter:
				my $majority_net = "crashtest_majority_".$fault_index;

				push(@file_additions, "wire ".$majority_net.";\n");
			
				push(@file_additions, "assign ".$majority_net." = \(crashtest_fv_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] \& ".$faulty_net_input_a."\) \| \(crashtest_fv_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] \& ".$faulty_net_input_b."\) \| \(".$faulty_net_input_a." \& ".$faulty_net_input_b."\);\n");

				push(@file_additions, "assign ".$faulty_net_output_a." = crashtest_fe_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] ? ".$majority_net." : ".$faulty_net_input_a.";\n");
				push(@file_additions, "assign ".$faulty_net_output_b." = crashtest_fe_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] ? ".$majority_net." : ".$faulty_net_input_b.";\n");
				
				$fault_index++;
			}



		} elsif($opt_string{path_delay} > 0) {
			my $num_fault_locations = $opt_string{path_delay};

			my %ff_list = &get_list_flip_flops();

			my @ff_faulty = &select_faults_from_list(\%ff_list, $num_fault_locations);
			
			print "Selected Faulty cells:\n" if($opt_verbose);

			# Compute the depth of the scan chains:
			my $number_scan_chains = $opt_string{scan_chains};
			my $number_ff_faulty = $#ff_faulty+1;
			my $depth_scan_chains = ceil($number_ff_faulty/$number_scan_chains) - 1;

			$total_faults_locations = $number_ff_faulty;

			push(@file_additions, "// ############################# CRASHTEST ###################################:\n");
			push(@file_additions, "// Declaration signals needed for Crashtest:\n");
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fe_sc_".$scan_chain_index.";\n");
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fv_sc_".$scan_chain_index.";\n");
			}
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "wire [".$depth_scan_chains.":0] crashtest_path_delay_".$scan_chain_index.";\n");
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_signal_delayed_".$scan_chain_index.";\n");
			}

			print "I will insert ".$number_scan_chains." scan chains each of them deep ".$depth_scan_chains."\n" if($opt_verbose);

			my $fault_index = 0;
			foreach my $ff_name (@ff_faulty) {

				my $scan_chain_selected = &get_scan_chain_for_fault($fault_index);
				my $scan_chain_depth = &get_depth_for_fault($fault_index);

				print $ff_name.":\n" if($opt_verbose);

				my $cell = $top_module->find_cell($ff_name);

				my $faulty_net = "";
				my $clock_net = "";
				my $cell_input_found = 0;
				my $cell_clock_found = 0;
				foreach my $cell_pin ($cell->pins) {
					my $cell_port = $cell_pin->port;
					if(defined($cell_port)) {
						if(($cell_port->direction eq "in")&&($cell_port->name eq $flip_flops_input_port{$cell->submodname})) {
							$faulty_net = $cell_pin->netname;
							$cell_input_found = 1;
						}
						if(($cell_port->direction eq "in")&&($cell_port->name eq $flip_flops_clock_port{$cell->submodname})) {
							$clock_net = $cell_pin->netname;
							$cell_clock_found = 1;
						}

						
					}
				}

				if($cell_input_found && $cell_clock_found) {

					my $faulty_net_input = "crashtest_path_delay_".$scan_chain_selected."\[".$scan_chain_depth."\]";
				
					push(@file_additions, "\n// Fault Number: $fault_index; Faulty Net: $faulty_net; Scan Chain: $scan_chain_selected; Depth: $scan_chain_depth\n");

					if($scan_chain_depth == 0) {
						push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."\[0\] <= crashtest_fe_data\[".$scan_chain_selected."\];\n");
						push(@file_additions, "end\n");
					} else {
						push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] <= crashtest_fe_sc_".$scan_chain_selected."\[".($scan_chain_depth-1)."\];\n");
						push(@file_additions, "end\n");
					}

					push(@file_additions, "always@(posedge ".$clock_net.")\n");
					push(@file_additions, "begin\n");
					push(@file_additions, "\tcrashtest_signal_delayed_".$scan_chain_selected."\[".$scan_chain_depth."\] <= $faulty_net;\n");
					push(@file_additions, "end\n");

					push(@file_additions, "assign ".$faulty_net_input." = crashtest_fe_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] ? crashtest_signal_delayed_".$scan_chain_selected."\[".$scan_chain_depth."\] : $faulty_net;\n");		

					my @cell_pin_list = $cell->pins_sorted;

					my %new_cell_ports = ();
					my %new_cell_pins = ();

					for(my $pin_index = 0; $pin_index <= $#cell_pin_list; $pin_index++) {
						my $cell_port = $cell_pin_list[$pin_index]->port;

						$new_cell_pins{$cell_port->name} = $cell_pin_list[$pin_index]->name;
						if($cell_port->name eq $flip_flops_input_port{$cell->submodname}) {
							# Change the signal with the new one
							$new_cell_ports{$cell_port->name} = $faulty_net_input;
							
						} else {
							$new_cell_ports{$cell_port->name} = $cell_pin_list[$pin_index]->netname;
						}
					}

					my $cell_name = $cell->name;
					my $cell_module_name = $cell->submodname;
					my $cell_module = $nl->find_module($cell_module_name);
					$cell->delete;

					my $new_cell_object = $top_module->new_cell(name=>$cell_name, submodname=>$cell_module_name,
						filename=>$top_module->filename, lineno=>$top_module->lineno);


					
					for my $key ( keys %new_cell_ports ) {
						my $cell_port = $cell_module->find_port($key);

						$new_cell_object->new_pin(name=>$new_cell_pins{$key}, netname=>$new_cell_ports{$key}, port=>$cell_port,
						filename=>$top_module->filename, lineno=>$top_module->lineno);
					}

					$new_cell_object->lint();


				} else {
					print "Error, I could not find the input for FF: ".$cell->name."\n";
					exit(1);
				}

				$fault_index++;
			}

		} elsif($opt_string{transient} > 0) {
			my $num_fault_locations = $opt_string{transient};

			my %ff_list = &get_list_flip_flops();

			my @ff_faulty = &select_faults_from_list(\%ff_list, $num_fault_locations);
			
			print "Selected Faulty cells:\n" if($opt_verbose);

			# Compute the depth of the scan chains:
			my $number_scan_chains = $opt_string{scan_chains};
			my $number_ff_faulty = $#ff_faulty+1;
			my $depth_scan_chains = ceil($number_ff_faulty/$number_scan_chains) - 1;

			$total_faults_locations = $number_ff_faulty;

			push(@file_additions, "// ############################# CRASHTEST ###################################:\n");
			push(@file_additions, "// Declaration signals needed for Crashtest:\n");
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fe_sc_".$scan_chain_index.";\n");
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fv_sc_".$scan_chain_index.";\n");
			}
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "wire [".$depth_scan_chains.":0] crashtest_transient_".$scan_chain_index.";\n");
			}

			push(@file_additions, "reg \[".($num_fault_locations-1).":0\] crashtest_transient_disabler;\n");

			print "I will insert ".$number_scan_chains." scan chains each of them deep ".$depth_scan_chains."\n" if($opt_verbose);

			my $fault_index = 0;
			foreach my $ff_name (@ff_faulty) {

				my $scan_chain_selected = &get_scan_chain_for_fault($fault_index);
				my $scan_chain_depth = &get_depth_for_fault($fault_index);

				print "Inject Transient Fault in cell ".$ff_name.":\n" if($opt_verbose);

				my $cell = $top_module->find_cell($ff_name);

				my $faulty_net = "";
				my $clock_net = "";
				my $cell_input_found = 0;
				my $cell_clock_found = 0;
				foreach my $cell_pin ($cell->pins) {
					my $cell_port = $cell_pin->port;
					if(defined($cell_port)) {
						if(($cell_port->direction eq "in")&&($cell_port->name eq $flip_flops_input_port{$cell->submodname})) {
							$faulty_net = $cell_pin->netname;
							$cell_input_found = 1;
						}
						if(($cell_port->direction eq "in")&&($cell_port->name eq $flip_flops_clock_port{$cell->submodname})) {
							$clock_net = $cell_pin->netname;
							$cell_clock_found = 1;
						}

						
					}
				}

				if($cell_input_found && $cell_clock_found) {

					my $faulty_net_input = "crashtest_transient_".$scan_chain_selected."\[".$scan_chain_depth."\]";
				
					push(@file_additions, "\n// Fault Number: $fault_index; Faulty Net: $faulty_net; Scan Chain: $scan_chain_selected; Depth: $scan_chain_depth\n");

					if($scan_chain_depth == 0) {
						push(@file_additions, "always\@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."\[0\] <= crashtest_fe_data\[".$scan_chain_selected."\];\n");
						push(@file_additions, "end\n");
					} else {
						push(@file_additions, "always\@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] <= crashtest_fe_sc_".$scan_chain_selected."\[".($scan_chain_depth-1)."\];\n");
						push(@file_additions, "end\n");
					}

					my $crashtest_disabler_name = "crashtest_transient_disabler\[".$fault_index."\]";

					
					push(@file_additions, "always\@(posedge crashtest_fe_clk or posedge ".$clock_net.")\n");
					push(@file_additions, "begin\n");
					push(@file_additions, "\tif(".$crashtest_disabler_name.")\n");
					push(@file_additions, "\tbegin\n");
					push(@file_additions, "\t\t".$crashtest_disabler_name." <= 0;\n");
					push(@file_additions, "\tend\n");
					push(@file_additions, "\tif(crashtest_fe_clk)\n");
					push(@file_additions, "\tbegin\n");
					push(@file_additions, "\t\t".$crashtest_disabler_name." <= 1;\n");
					push(@file_additions, "\tend\n");
					push(@file_additions, "end\n");
					push(@file_additions, "assign ".$faulty_net_input." = crashtest_fe_sc_".$scan_chain_selected."\[".$scan_chain_depth."\] && ".$crashtest_disabler_name." ? ~".$faulty_net." : $faulty_net;\n");		

					my @cell_pin_list = $cell->pins_sorted;

					my %new_cell_ports = ();
					my %new_cell_pins = ();

					for(my $pin_index = 0; $pin_index <= $#cell_pin_list; $pin_index++) {
						my $cell_port = $cell_pin_list[$pin_index]->port;

						$new_cell_pins{$cell_port->name} = $cell_pin_list[$pin_index]->name;
						if($cell_port->name eq $flip_flops_input_port{$cell->submodname}) {
							# Change the signal with the new one
							$new_cell_ports{$cell_port->name} = $faulty_net_input;
							
						} else {
							$new_cell_ports{$cell_port->name} = $cell_pin_list[$pin_index]->netname;
						}
					}

					my $cell_name = $cell->name;
					my $cell_module_name = $cell->submodname;
					my $cell_module = $nl->find_module($cell_module_name);
					$cell->delete;

					my $new_cell_object = $top_module->new_cell(name=>$cell_name, submodname=>$cell_module_name,
						filename=>$top_module->filename, lineno=>$top_module->lineno);


					
					for my $key ( keys %new_cell_ports ) {
						my $cell_port = $cell_module->find_port($key);

						$new_cell_object->new_pin(name=>$new_cell_pins{$key}, netname=>$new_cell_ports{$key}, port=>$cell_port,
						filename=>$top_module->filename, lineno=>$top_module->lineno);
					}

					$new_cell_object->lint();


				} else {
					print "Error, I could not find the input for FF: ".$cell->name."\n";
					exit(1);
				}

				$fault_index++;
			}

		} elsif($opt_string{stuck_open} > 0) {
			my $num_fault_locations = $opt_string{stuck_open};

			# Build the list of nodes where a stuck-at fault can be injected. To each net is attached the probability that 
			# a fault hits the network. 

			my %node_list = &return_valid_nets_for_faults();

			if($opt_verbose) {
				print "Nets where I can insert stuck-open faults:\n";
				for my $net ( keys %node_list ) {
					my $fault_probability = $node_list{$net};
					print "$net => $fault_probability\n";
				}
			}

			# Call the function to select which faulty nets will be instrumented
			my @faulty_nets = &select_faults_from_list(\%node_list, $num_fault_locations);

			if($opt_verbose) {
				print "Nets selected to inject stuck-open faults:\n";
				foreach (@faulty_nets) {
					print $_."\n";
				}
			}

			# Compute the depth of the scan chains:
			my $number_scan_chains = $opt_string{scan_chains};
			my $number_faulty_nets = $#faulty_nets+1;
			my $depth_scan_chains = ceil($number_faulty_nets/$number_scan_chains) - 1;

			$total_faults_locations = $number_faulty_nets;

			push(@file_additions, "// ############################# CRASHTEST ###################################:\n");
			push(@file_additions, "// Declaration signals needed for Crashtest:\n");
			for(my $scan_chain_index = 0; $scan_chain_index < $number_scan_chains; $scan_chain_index++) {
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fe_sc_".$scan_chain_index.";\n");
				push(@file_additions, "reg [".$depth_scan_chains.":0] crashtest_fv_sc_".$scan_chain_index.";\n");
			}

			
			# Select 8 signals from the netlist
			my @random_nets = &select_faults_from_list(\%node_list, 8);

			push(@file_additions, "wire	random_out;\n");
			push(@file_additions, "// Instantiate the lfsr that will be used as random generator\n");
			push(@file_additions, "lfsr random_gen (\n");
			push(@file_additions, "  .data({");

			if($#random_nets < 7) {
				for(my $i = $#random_nets-1; $i <= 7; $i++) {
					$random_nets[$i] = "1'b0";
				}
			}
			for(my $i = 0; $i < 8; $i++) {
				push(@file_additions, $random_nets[$i]);
				if($i < 7) {
					push(@file_additions, ",");					
				}
			}

			push(@file_additions, "}),\n");

			# Select reset signals from the netlist
			my @random_reset_net = &select_faults_from_list(\%node_list, 1);
			if($#random_reset_net < 0) {
				@random_reset_net = "1'b1";
			}

			push(@file_additions, "  .reset(".$random_reset_net[0]."),\n");
			push(@file_additions, "  .out(random_out),\n");
			push(@file_additions, "  .enable(1'b1),\n");
			# Use the first clock network to run the lfsr
			if($#clock_nets < 0) {
				print STDERR "You need to define at least one clock network to inject stuck_open faults.\n";
				exit(1);
			}
			push(@file_additions, "  .clk(".$clock_nets[0].") );\n");

			print "I will insert ".$number_scan_chains." scan chains each of them deep ".$depth_scan_chains."\n" if($opt_verbose);


			my $fault_index = 0;
			print "Nets selected to inject stuck-at faults:\n" if($opt_verbose);
			foreach (@faulty_nets) {

				my $faulty_net = $_;

				my $scan_chain_selected = &get_scan_chain_for_fault($fault_index);
				my $scan_chain_depth = &get_depth_for_fault($fault_index);
				
				# If the net is an input change it right away, if the net is output of a gate change the output
				my $name_signal = $faulty_net;
				my $index_signal = -1;
				
				if($faulty_net =~ m/(.*)\[(\d*)\]/) {
					$name_signal = $1;
					$index_signal = $2;
				}

				$name_signal =~ s/\\//g;
				$name_signal =~ s/\//\_/g;

				print "Faulty net selected: ".$faulty_net."\n" if($opt_verbose);
				my $driver_found = 0;
				my $original_net = $top_module->find_net($faulty_net);

				# Extract from which cell the signal is generated:					
				foreach my $cell ($top_module->cells) {
					foreach my $cell_pin ($cell->pins) {
						my $cell_port = $cell_pin->port;
						if(defined($cell_port)) {
							if(($cell_port->direction eq "out")&&($cell_pin->netname eq $faulty_net)) {
								print $cell->dump if($opt_verbose);
								print "Found Driver cell ".$cell->name." for selected faulty network ".$faulty_net."\n"  if($opt_verbose);
								$driver_found = 1;

								my $faulty_net_input = "crashtest_stuck_open_".&trim($name_signal)."_".$fault_index;
								if($index_signal > 0) {
									 $faulty_net_input = "crashtest_stuck_open_".$name_signal."_".$index_signal."_".$fault_index;
								}

								$top_module->new_net(name=>$faulty_net_input,
									data_type=>"wire", 
									comment=>"// Crashtest: Wire needed to insert fault at the output of a gate",);

								push(@file_additions, "\n// Fault Number: $fault_index; Faulty Net: $faulty_net; Scan Chain: $scan_chain_selected; Depth: $scan_chain_depth\n");

								if($scan_chain_depth == 0) {
									push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
									push(@file_additions, "begin\n");
									push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[0] <= crashtest_fe_data[".$scan_chain_selected."];\n");
									push(@file_additions, "end\n");
								} else {
									push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
									push(@file_additions, "begin\n");
									push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= crashtest_fe_sc_".$scan_chain_selected."[".($scan_chain_depth-1)."];\n");
									push(@file_additions, "end\n");
								}

								push(@file_additions, "always@(".$clock_nets[0].")\n");
								push(@file_additions, "begin\n");
								push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= random_out;\n");
								push(@file_additions, "end\n");

								push(@file_additions, "assign ".$faulty_net." = crashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] ? crashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] : $faulty_net_input;\n");

								my @cell_pin_list = $cell->pins_sorted;

								my %new_cell_ports = ();
								my %new_cell_pins = ();

								
								for(my $pin_index = 0; $pin_index <= $#cell_pin_list; $pin_index++) {
									my $cell_port = $cell_pin_list[$pin_index]->port;

									$new_cell_pins{$cell_port->name} = $cell_pin_list[$pin_index]->name;
									if($cell_pin->name eq $cell_port->name) {
										# Change the signal with the new one
										$new_cell_ports{$cell_port->name} = $faulty_net_input;
										
									} else {
										$new_cell_ports{$cell_port->name} = $cell_pin_list[$pin_index]->netname;
									}
								}

								my $cell_name = $cell->name;
								my $cell_module_name = $cell->submodname;
								my $cell_module = $nl->find_module($cell_module_name);
								$cell->delete;

								my $new_cell_object = $top_module->new_cell(name=>$cell_name, submodname=>$cell_module_name,
									filename=>$top_module->filename, lineno=>$top_module->lineno);


								
								for my $key ( keys %new_cell_ports ) {
									my $cell_port = $cell_module->find_port($key);

									$new_cell_object->new_pin(name=>$new_cell_pins{$key}, netname=>$new_cell_ports{$key}, port=>$cell_port,
									filename=>$top_module->filename, lineno=>$top_module->lineno);
								}

								$new_cell_object->lint();

							}
						}
					}
				}

				if(!$driver_found) {
					# If the signal is not generated from a cell, look if the signal is an input of the top module:
					$driver_found = 1;

					my $faulty_net_input = "crashtest_".&trim($name_signal)."_".$fault_index;
					if($index_signal > 0) {
						 $faulty_net_input = "crashtest_".&trim($name_signal)."_".$index_signal."_".$fault_index;
					}

					# If the net is a primary input, I will need to change all the connections to the port: create a new wire to connect the mux to the gates that use it as input
										
					$top_module->new_net(name=>$faulty_net_input,
						data_type=>"wire", 
						comment=>"// Crashtest: Wire needed to insert fault in input signals",);

					push(@file_additions, "\n// Fault Number: $fault_index; Faulty Net: $faulty_net; Scan Chain: $scan_chain_selected; Depth: $scan_chain_depth\n");
				
					if($scan_chain_depth == 0) {
						push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[0] <= crashtest_fe_data[".$scan_chain_selected."];\n");
						push(@file_additions, "end\n");

					} else {
						push(@file_additions, "always@(posedge crashtest_fe_clk)\n");
						push(@file_additions, "begin\n");
						push(@file_additions, "\tcrashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= crashtest_fe_sc_".$scan_chain_selected."[".($scan_chain_depth-1)."];\n");
						push(@file_additions, "end\n");
					}

					push(@file_additions, "always@(*)\n");
					push(@file_additions, "begin\n");
					push(@file_additions, "\tcrashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] <= crashtest_fv_data[".$scan_chain_selected."];\n");
					push(@file_additions, "end\n");

					push(@file_additions, "assign ".$faulty_net_input." = crashtest_fe_sc_".$scan_chain_selected."[".$scan_chain_depth."] ? crashtest_fv_sc_".$scan_chain_selected."[".$scan_chain_depth."] : ".$faulty_net.";\n");

					# Find all the cells that use the original signal as input and change their connections	
					foreach my $cell ($top_module->cells) {
						foreach my $cell_pin ($cell->pins) {
							my $cell_port = $cell_pin->port;
							if(defined($cell_port)) {
								if(($cell_port->direction eq "in")&&($cell_pin->netname eq $faulty_net)) {
									print $cell->dump if($opt_verbose);
									print "Found cell ".$cell->name." that uses ".$faulty_net." as input.\n" if($opt_verbose);

									my @cell_pin_list = $cell->pins_sorted;

									my %new_cell_ports = ();
									my %new_cell_pins = ();
									for(my $pin_index = 0; $pin_index <= $#cell_pin_list; $pin_index++) {
										my $cell_port = $cell_pin_list[$pin_index]->port;

										$new_cell_pins{$cell_port->name} = $cell_pin_list[$pin_index]->name;
										if($cell_pin->name eq $cell_port->name) {
											# Change the signal with the new one
											$new_cell_ports{$cell_port->name} = $faulty_net_input;
										
										} else {
											$new_cell_ports{$cell_port->name} = $cell_pin_list[$pin_index]->netname;
										}
										if($pin_index < $#cell_pin_list) {

										}
									}

									my $cell_name = $cell->name;
									my $cell_module_name = $cell->submodname;
									my $cell_module = $nl->find_module($cell_module_name);
									$cell->delete;

									my $new_cell_object = $top_module->new_cell(name=>$cell_name, submodname=>$cell_module_name,
										filename=>$top_module->filename, lineno=>$top_module->lineno);


								
									for my $key ( keys %new_cell_ports ) {
										my $cell_port = $cell_module->find_port($key);

										$new_cell_object->new_pin(name=>$new_cell_pins{$key}, netname=>$new_cell_ports{$key}, port=>$cell_port,
										filename=>$top_module->filename, lineno=>$top_module->lineno);
									}

								
								
									$new_cell_object->lint();

								}
							}
						}
					}

					# I need to change all the assign statements to make them use the modified net:
					print "Analyze statements:\n" if($opt_verbose);
					foreach my $statement ($top_module->statements_sorted) {
						print "LHS: ".$statement->lhs if($opt_verbose);
						print " RHS: ".$statement->rhs."\n" if($opt_verbose);

						if($statement->rhs eq $faulty_net) {
							$assign_rhs{$statement->rhs} = $faulty_net_input;
							print "Substituition ".$statement->rhs." to ".$faulty_net_input." annotated!\n" if($opt_verbose);
						}
					}
				}

				$fault_index++;
			}

		} else {
			print "No faults will be injected.\n";
			exit(0);
		}

		&add_crashtest_inputs();
		
		# Write modified netlist:

		my $output_filename = "";
		if($opt_string{output_filename} eq "") {
			$top_module->name.".".&format_time().".ct.v";
		} else {
			$output_filename = $opt_string{output_filename};
		}

		my $modified_netlist_file = $opt_string{output_directory}."/".$output_filename;
	
		open(OUTPUT_NETLIST, ">".$modified_netlist_file) or die $!;

		print OUTPUT_NETLIST
		"// ========== Copyright Header Begin ==========================================\n",
		"// \n",
		"// CrashTest: Fast and Accurate Fault Analysis Platform\n",
		"// Copyright (C) 2008-2010  University of Michigan\n",
		"// http://www.eecs.umich.edu/crashtest\n",
		"// Contributors include Andrea Pellegrini, Valeria Bertacco, Todd Austin, \n",
		"// Kypros Constantinides ,Junhao Jiang, Shobana Sudhakar, and Dan Zhang\n",
		"// \n",
		"// The above named program is free software; you can redistribute it and/or\n",
		"// modify it under the terms of the GNU General Public\n",
		"// License version 2 as published by the Free Software Foundation.\n",
		"// \n",
		"// The above named program is distributed in the hope that it will be \n",
		"// useful, but WITHOUT ANY WARRANTY; without even the implied warranty of\n",
		"// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU\n",
		"// General Public License for more details.\n",
		"// \n",
		"// You should have received a copy of the GNU General Public\n",
		"// License along with this work; if not, write to the Free Software\n",
		"// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.\n",
		"//\n",
		"// Andrea Pellegrini <apellegr@umich.edu>\n",
		"// Valeria Bertacco  <valeria@umich.edu>\n",
		"// Todd Austin       <austin@umich.edu>\n",
		"// University of Michigan\n",
		"// Electrical Engineering and Computer Science Dept.\n",
		"// 2260 Hayward St.\n",
		"// Ann Arbor, MI 48109-2121\n",
		"// \n",
		"// ========== Copyright Header End ============================================\n\n";

		print OUTPUT_NETLIST
		"/********************************************************************************/\n",
		"// Copyright Andrea Pellegrini, University of Michigan, 2010\n",
		"// File Generated through the CrashTest suite.\n",
		"// Invocation Paramters:\n";

		for my $key (keys %opt_string) {
			print OUTPUT_NETLIST "// ".$key."=>".$opt_string{$key}."\n";
		}

		print OUTPUT_NETLIST "\n// ------------------------------------------------------------------\n";
		print OUTPUT_NETLIST "// I was able to inject faults in $total_faults_locations locations\n";
		print OUTPUT_NETLIST "// ------------------------------------------------------------------\n\n";
		print OUTPUT_NETLIST "/********************************************************************************/\n";


		print OUTPUT_NETLIST "//-----------------------------------------------------\n";
		print OUTPUT_NETLIST "// Design Name : lfsr\n";
		print OUTPUT_NETLIST "// From        : http://www.asic-world.com/code/hdl_models/lfsr.v\n";
		print OUTPUT_NETLIST "// Function    : Linear feedback shift register\n";
		print OUTPUT_NETLIST "// Coder       : Deepak Kumar Tala\n";
		print OUTPUT_NETLIST "//-----------------------------------------------------\n";
		print OUTPUT_NETLIST "module lfsr    (\n";
		print OUTPUT_NETLIST "data            ,  // Input of the counter\n";
		print OUTPUT_NETLIST "out             ,  // Output of the counter\n";
		print OUTPUT_NETLIST "enable          ,  // Enable  for counter\n";
		print OUTPUT_NETLIST "clk             ,  // clock input\n";
		print OUTPUT_NETLIST "reset              // reset input\n";
		print OUTPUT_NETLIST ");\n";
		print OUTPUT_NETLIST "//----------Output Ports--------------\n";
		print OUTPUT_NETLIST "output [7:0] out;\n";
		print OUTPUT_NETLIST "//------------Input Ports--------------\n";
		print OUTPUT_NETLIST "input [7:0] data;\n";
		print OUTPUT_NETLIST "input enable, clk, reset;\n";
		print OUTPUT_NETLIST "//------------Internal Variables--------\n";
		print OUTPUT_NETLIST "reg [7:0] out;\n";
		print OUTPUT_NETLIST "wire        linear_feedback;\n";
		print OUTPUT_NETLIST "//-------------Code Starts Here-------\n";
		print OUTPUT_NETLIST "assign linear_feedback = !(out[7] ^ out[3]);\n";
		print OUTPUT_NETLIST "always @(posedge clk)\n";
		print OUTPUT_NETLIST "begin\n";
		print OUTPUT_NETLIST "  if (reset) begin // active high reset\n";
		print OUTPUT_NETLIST "    out <= 8'b0 ;\n";
		print OUTPUT_NETLIST "  end else if (enable)\n";
		print OUTPUT_NETLIST "  begin\n";
		print OUTPUT_NETLIST "    out <= {out[6],out[5],\n";
		print OUTPUT_NETLIST "            out[4],out[3],\n";
		print OUTPUT_NETLIST "            out[2],out[1],\n";
		print OUTPUT_NETLIST "            out[0], linear_feedback};\n";
		print OUTPUT_NETLIST "  end\n";
		print OUTPUT_NETLIST "end\n";
		print OUTPUT_NETLIST "endmodule // End Of Module counter\n";

		my @file_to_update = $top_module->verilog_text;
		
		my @output_netlist = ();

		# Use @file_additions to put the new lines in the file, before endmodule


		my $assign_flag = 0;
		my $rhs_flag = 0;
		foreach(@file_to_update) {
			my $file_line = $_;
			if($file_line) {
				print "File line: ".$file_line."\n" if($opt_verbose);
				if($file_line =~ m/endmodule/) {
				
					foreach(@file_additions) {
					
						my $additions_line = $_;
						# Fix signals name issues with Synplicity 
						# Change signals names from \p0/c0mmu/icache0/rl_LRU_148_0 p0_c0mmu_icache0_rl_LRU_148_0
						$additions_line =~ s/\/\//comment\_\_comment/g;
						$additions_line =~ s/\\(\w*)/$1/g;
						$additions_line =~ s/(\w*)\/(\w*)/$1\_$2/g;
						$additions_line =~ s/\*Logic0\*/Logic0/g;
						$additions_line =~ s/\*Logic1\*/Logic1/g;
						$additions_line =~ s/comment\_\_comment/\/\//g;
					
						push(@output_netlist, $additions_line);
					}
				} elsif($file_line =~ m/\s*assign\s*/) {

# Should be done like this:
#	When you have a ContAssign object, just change $obj->lhs("new expression")
#	then write as normal with the verilog_text function.
#
#	-Wilson


					print "Found Assign!\n" if($opt_verbose);
					$assign_flag = 1;
				} elsif($file_line =~ m/\s*=\s*/ && $assign_flag) {
					print "Start RHS!\n" if($opt_verbose);
					$rhs_flag = 1;
				} elsif($assign_flag && $rhs_flag) {
					if($file_line =~ m/\s*;\s*/) {
						# The assign statement is terminated
						$assign_flag = 0;
						$rhs_flag = 0;
						print "Close Assign!\n" if($opt_verbose);
					} else {
						for my $key (keys %assign_rhs) {
							if($file_line eq $key) {
								$file_line = $assign_rhs{$key};
								print "Change ".$key." to ".$assign_rhs{$key}."\n" if($opt_verbose);
							}
						}
					}
					
				}
				
				# Fix signals name issues with Synplicity 
				# Change signals names from \p0/c0mmu/icache0/rl_LRU_148_0 p0_c0mmu_icache0_rl_LRU_148_0
				$file_line =~ s/\/\//comment\_\_comment/g;
				$file_line =~ s/\\(\w*)/$1/g;
				$file_line =~ s/(\w*)\/(\w*)/$1\_$2/g;
				$file_line =~ s/\*Logic0\*/Logic0/g;
				$file_line =~ s/\*Logic1\*/Logic1/g;
				$file_line =~ s/comment\_\_comment/\/\//g;
				
				push(@output_netlist, $file_line);
			}
		}

		print OUTPUT_NETLIST @output_netlist;
		close OUTPUT_NETLIST;

	}
}


exit(0);

### Functions ##################################################################
sub usage() {
	print basename($0) . " --top_module=[name module to inject] --input_netlist=[path of the input netlist] --library=[path of the library used] --scan_chains=[number of scan chains] --clock_net=[clock signal name] [--fault_type=num_faults ...]\n\n";
	print "\tTypes of faults supported: stuck_at, bridge, path_delay, transient, stuck_open\n\n";
	print "\tFor each fault type indicate how many fault locations will be inserted:\n";
	print "\tfor example, to inject 1024 stuck-at faults, the argument set should be: --stuck_at=1024\n";
	print "\tOnly one type of fault will be injected at the time!\n";
	print "\tBridging fault types are generated randomly only for the supported gates.\n";

	print "\tDefault Options:\n";	
	foreach my $opt (sort keys %opt_string) {
		print "\t\t--$opt='$opt_string{$opt}'\n";
	}
	print "\t--help    for Help\n";
	print "\t--verbose for Verbose mode\n";

	exit(0);
}

sub can_read_file($$) {
	my $file_name = $_[0];

	if (!(-e $file_name)) { 
		print STDERR "The file ".$file_name." does not exists.\n";
		exit(1);
	}	
	if (!(-r $file_name)) { 
		print STDERR "The file ".$file_name." cannot be read.\n";
		exit(1);
	}
}

sub can_write_file($$) {
	my $file_name = $_[0];

	if (!(-w $file_name)) { 
		print STDERR "The file ".$file_name." cannot be written.\n";
		exit(1);
	}
}

sub add_crashtest_inputs() {
	# Add the extra signals needed for CrashTest

	foreach my $top_module ($nl->top_modules_sorted) {
		if($top_module->name eq $opt_string{top_module}) {

			# Create port for FaultEnable Data:
			my $fe_data_port = $top_module->new_port (name=>"crashtest_fe_data",
				filename=>$top_module->filename, lineno=>$top_module->lineno);

			# Create declaration for FaultEnable Data:
			my $fe_data_net = $top_module->new_net (name=>"crashtest_fe_data",
				#filename=>$top_module->filename, lineno=>$top_module->lineno,
				direction=>"in", decl_type=>"input [".($opt_string{scan_chains}-1).":0]",
				data_type=>"",
				width=>$opt_string{scan_chains}, 
				comment=>"// Crashtest: Fault Enable Data from wrapper",);
	

			# Create port for FaultEnable Clock:
			my $fe_clock_port = $top_module->new_port (name=>"crashtest_fe_clk",
				filename=>$top_module->filename, lineno=>$top_module->lineno);

			# Create declaration for FaultEnable Clock:
			my $fe_clock_net = $top_module->new_net (name=>"crashtest_fe_clk",
				#filename=>$top_module->filename, lineno=>$top_module->lineno,
				direction=>"in", decl_type=>"input",
				data_type=>"",
				width=>$opt_string{scan_chains}, 
				comment=>"// Crashtest: Fault Enable Clock from wrapper",);

			# Create port for FaultValue Data:
			my $fv_data_port = $top_module->new_port (name=>"crashtest_fv_data",
				filename=>$top_module->filename, lineno=>$top_module->lineno);

			# Create declaration for FaultValue Data:
			my $fv_data_net = $top_module->new_net (name=>"crashtest_fv_data",
				#filename=>$top_module->filename, lineno=>$top_module->lineno,
				direction=>"in", decl_type=>"input [".($opt_string{scan_chains}-1).":0]",
				data_type=>"",
				width=>$opt_string{scan_chains}, 
				comment=>"// Crashtest: Fault Value Data from wrapper",);
	

			# Create port for FaultValue Clock:
			my $fv_clock_port = $top_module->new_port (name=>"crashtest_fv_clk",
				filename=>$top_module->filename, lineno=>$top_module->lineno);

			# Create declaration for FaultValue Clock:
			my $fv_clock_net = $top_module->new_net (name=>"crashtest_fv_clk",
				#filename=>$top_module->filename, lineno=>$top_module->lineno,
				direction=>"in", decl_type=>"input",
				data_type=>"",
				width=>$opt_string{scan_chains}, 
				comment=>"// Crashtest: Fault Value Clock from wrapper",);

			
		}
	}	
}


### Given a list of possible faults and weights, returns a given number of random elements ########################################################
# Input 1: hashmap containing Element=>Weight
# Input 2: Number of random Elements to return
# Output: List of Elements returned

sub select_faults_from_list() {
	my(%input_list) = %{$_[0]};
	my($fault_locations) = $_[1];

	my @result_list = ();

	my $list_weight = 0;
	my $possible_faults = 0;
	for my $key ( keys %input_list ) {
		my $fault_probability = $input_list{$key};
		$list_weight += $fault_probability;
		$possible_faults++;
	}

	print "Num selectable faults: ".$possible_faults."\n" if($opt_verbose);

	if($possible_faults <= $fault_locations) {
		print "Num selectable faults less than number of faults requested, all will be returned.\n" if($opt_verbose);
		for my $key ( keys %input_list ) {
			push(@result_list, $key);
		}
		return @result_list;
	}
	
	my %temporary_result_list = ();

	my $num_selected_faults = 0;
	while($num_selected_faults < $fault_locations) {
		my $random_number = rand($list_weight);
		print "Random number selected: ".$random_number."/".$list_weight."\n" if($opt_verbose);
		for my $key ( keys %input_list ) {
			if($random_number > 0) {
				my $fault_probability = $input_list{$key};
				$random_number -= $fault_probability;
				if(($random_number <= 0)&&(! exists($temporary_result_list{$key}))) {
					# Add the fault in the list of selected faults
					print "Selected fault: ".$key."\n" if($opt_verbose);
					$temporary_result_list{$key} = 1;
					$num_selected_faults++;
				}
			}
		}
	}

	for my $key ( keys %temporary_result_list ) {
		push(@result_list, $key);
	}
	return @result_list;
}


### Given the number of a fault computes the scan chain it is located ########################################################
# Input : Fault indexq
# Output : Number of scan chain

sub get_scan_chain_for_fault() {
	my($fault_location) = $_[0];

	return $fault_location % $opt_string{scan_chains};
}

### Given the number of a fault computes the depth it is located in the scan chain ########################################################
# Input : Fault index
# Output : Depth in the scan chain

sub get_depth_for_fault() {
	my($fault_location) = $_[0];

	return floor($fault_location / $opt_string{scan_chains});
}

### Returns the list of names of ff cells in the design ########################################################
# Output : Names of list of ff cells in the top_module, a value with area information is attached to each element
sub get_list_flip_flops() {
	print "Extract Flip Flops cells:\n" if($opt_verbose);

	my %ff_list = ();

	foreach my $top_module ($nl->top_modules_sorted) {
		if($top_module->name eq $opt_string{top_module}) {
			foreach my $cell ($top_module->cells) {
				foreach my $ff_name (@flip_flops_name_filter) {
					if($ff_name eq $cell->submodname) {
						# Same area for each:
						my $cell_name = $cell->name;
						$ff_list{$cell_name} = 1;
					}
				}
			}
		}

	}

	return %ff_list;
}

### Returns the list of names of cells in the design with at least two input ports ########################################################
# Output : Names of list of ff cells in the top_module, a value with area information is attached to each element
sub get_list_gates_at_least_two_inputs() {
	print "Extract cells with at least two inputs:\n" if($opt_verbose);

	my %cell_list = ();

	foreach my $top_module ($nl->top_modules_sorted) {
		if($top_module->name eq $opt_string{top_module}) {
			foreach my $cell ($top_module->cells) {
				my $num_input = 0;
				foreach my $cell_pin ($cell->pins) {
					my $cell_port = $cell_pin->port;
					if(defined($cell_port)) {
						my $net_part_of_clock_tree = 0;
						if(($cell_port->direction eq "in")) {
							foreach(@clock_tree_nets) {
								if($cell_pin->netname eq $_) {
									$net_part_of_clock_tree = 1;
								}
							}
							if(!$net_part_of_clock_tree) {
								$num_input++;
							}
						}
					}
				}
				if($num_input >= 2) {
					my $cell_name = $cell->name;
					# Same area for each:
					$cell_list{$cell_name} = 1;
				}
			}
		}
	}
	return %cell_list;
}

### Returns date and time in a format uasble in the file system ########################################################
# Output :  returns date and time in the format YYYY-MM-DD_hh.mm.ss
sub format_time() {
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
  my $time = sprintf ("%4d-%02d-%02d_%02d.%02d.%02d",
                      $year+1900,$mon+1,$mday,$hour,$min,$sec);
  return $time;
}

### Returns all the nets that are not in the clock tree ########################################################
# Output : hash where the key is the name of the net and the value is the area (for gtech they are all equal)
sub return_valid_nets_for_faults() {

	print "Begin list of networks where I can inject faults:\n" if($opt_verbose);

	my %node_list = ();

	my %clock_tree_hash = ();
	foreach(@clock_tree_nets) {
		$clock_tree_hash{$_} = 1;
	}

	foreach my $top_module ($nl->top_modules_sorted) {
		if($top_module->name eq $opt_string{top_module}) {

			print "Analyze top module: ".$top_module->name.":\n";

			# Let's select only networks that are input to some cell but are
			# are not in the clock tree

			# I have no area information, all nets will have the same area
			my $net_fault_probability = 1;

			foreach my $cell ($top_module->cells) {

				foreach my $cell_pin ($cell->pins) {
					my $cell_port = $cell_pin->port;

					if(defined($cell_pin->net)) {
						my $is_valid_signal = 0;
						my $net_part_of_clock_tree = 0;

						my $net = $cell_pin->net;
						if(exists($clock_tree_hash{$cell_pin->netname})) {
							$net_part_of_clock_tree = 1;
						}
						if(defined($cell_port)) {
							if($cell_port->direction eq "in") {
								$is_valid_signal = 1;
							}
						}
				
						if(!$net_part_of_clock_tree && $is_valid_signal) {
							if(defined($net->msb) && defined($net->lsb)) {

								if($net->msb < $net->lsb) { 
									for(my $bit_index = $net->msb; $bit_index <= $net->lsb; $bit_index++) {
										print $net->name."\[".$bit_index."\]\n" if($opt_verbose);
										$node_list{$net->name."\[".$bit_index."\]"} = $net_fault_probability;
									}
								} else {
									for(my $bit_index = $net->lsb; $bit_index <= $net->msb; $bit_index++) {
										print $net->name."\[".$bit_index."\]\n" if($opt_verbose);
										$node_list{$net->name."\[".$bit_index."\]"} = $net_fault_probability;
									}
								}
							} else {
								print $net->name."\n" if($opt_verbose);
								$node_list{$net->name} = $net_fault_probability;
							}
						}
					}
				}
			}
		}
	}

	print "End list of networks where I can inject faults.\n" if($opt_verbose);
	return %node_list;
}




## Perl trim function to remove whitespace from the start and end of the string
sub trim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}
